from __future__ import print_function
import os
import subprocess
import threading
import time
import sys
import ctypes
from multiprocessing import Pool, Manager
os.system('color')

# IMPORTANT: Enter your path to Utilities24 folder. DO NOT add an extra backslash after the "\Utilities24"!
path_parameter_files = r"C:\Software\AutoDockTools"

# Construct the full path to the dpf4.py and gpf4.py script path
path_gpf = "{}{}".format(path_parameter_files, r"\Lib\site-packages\AutoDockTools\Utilities24\prepare_gpf4.py")
path_dpf = "{}{}".format(path_parameter_files, r"\Lib\site-packages\AutoDockTools\Utilities24\prepare_dpf4.py")

# Define color codes
HEADER = '\033[95m'
OKBLUE = '\033[94m'
OKCYAN = '\033[96m'
OKGREEN = '\033[92m'
WARNING = '\033[93m'
FAIL = '\033[91m'
ENDC = '\033[0m'
BOLD = '\033[1m'
UNDERLINE = '\033[4m'

# Ask the user for the interval in minutes
interval_seconds = 180 # Placeholder for global variable
def userInterval():
    global interval_seconds
    try:
        interval_minutes = float(raw_input("""
        Enter the interval in between the waking (i.e. combination tab/window switch + 'Right_arrow'key) of the script. 
        If you during this docking will be using your computer, then a short interval, i.e. 1 min, might be distracting.
        The wake-up interval is only needed for large screening of many ligands, but it is generally advisable to keep 
        in order to avoid any possible script-freezes. Default value is therefore 3 minute.
        
        Enter the number of minutes: """))
        interval_seconds = interval_minutes * 60  # Convert to seconds
    except ValueError:
        print("Invalid input. Using the default interval of 3 minute.")
        interval_seconds = 180  # Default to 3 minute if the input is invalid
    print("\n\n\n") # Prints spacing

# Constants for window management
SW_SHOW = 5  # Show the window

# Constants from the Windows API
KEYEVENTF_KEYDOWN = 0x0000
KEYEVENTF_KEYUP = 0x0002
VK_ESCAPE = 0x1B  # Virtual-key code for the "Escape" key
VK_RIGHT = 0x27  # Virtual-key code for the "Right Arrow" key
VK_RETURN = 0x0D  # Virtual-key code for the "Enter" key

def press_key():
    # Simulate key press
    ctypes.windll.user32.keybd_event(VK_RIGHT, 0, KEYEVENTF_KEYDOWN, 0)
    time.sleep(0.05)  # Short delay to simulate real key press
    # Simulate key release
    ctypes.windll.user32.keybd_event(VK_RIGHT, 0, KEYEVENTF_KEYUP, 0)

def get_console_window():
    # Get the handle of the console window
    return ctypes.windll.kernel32.GetConsoleWindow()

def bring_console_to_foreground(hwnd):
    # Show the window in case it's minimized
    ctypes.windll.user32.ShowWindow(hwnd, SW_SHOW)
    # Bring the window to the foreground
    ctypes.windll.user32.SetForegroundWindow(hwnd)

def keep_cmd_active(interval):
    hwnd = get_console_window()
    try:
        while True:
            if hwnd:  # Ensure we have a valid handle
                bring_console_to_foreground(hwnd)
                time.sleep(0.1) # Allow time to focus
                bring_console_to_foreground(hwnd)
                time.sleep(0.1) # Allow time to focus
                press_key()
                print("{}Awoke the script/cmd prompt{}".format(BOLD, ENDC))
            else:
                print("No console window handle found.")
            time.sleep(interval)  # Wait for the specified interval
    except KeyboardInterrupt:
        print("Stopped by user")

def title():
    # Define ASCII art
    ascii_art_title = """
                            __    _       __    __        _             ___         __        ____             __  
                           / /   (_)___ _/ /_  / /_____  (_)___  ____ _/   | __  __/ /_____  / __ \____  _____/ /__
                          / /   / / __ `/ __ \/ __/ __ \/ / __ \/ __ `/ /| |/ / / / __/ __ \/ / / / __ \/ ___/ //_/
                         / /___/ / /_/ / / / / /_/ / / / / / / / /_/ / ___ / /_/ / /_/ /_/ / /_/ / /_/ / /__/ ,<   
                        /_____/_/\__, /_/ /_/\__/_/ /_/_/_/ /_/\__, /_/  |_\__,_/\__/\____/_____/\____/\___/_/|_|  
                                /____/                        /____/                                             
    """
    credits = """
                                    ------------------------------------------------------------------
                                                            LightningAutoDock
                                                              by PyroTronix
                                                                  2024
                                            --------------------------------------------------



















    """

    # Print ASCII art with colors
    print(HEADER + BOLD + ascii_art_title + ENDC)
    print(OKCYAN + credits + ENDC)

# Set the timing for status prompt
status_interval = 180 # 180 sec per cmd prompt update - 3 minutes

def print_status(start_time, total_ligands, completed_dockings):
    while True:
        elapsed_time = time.time() - start_time
        elapsed_minutes = int(elapsed_time // 60)
        complete = completed_dockings.value
        print("{}Status: {} minutes have elapsed and {}/{} dockings have been completed.{}".format(BOLD, elapsed_minutes, complete, total_ligands, ENDC))
        sys.stdout.flush()  # Ensure it gets printed immediately
        time.sleep(status_interval)  # Sleep for "status_interval" seconds (status_interval/60 minutes)

def read_config(config_file='config.txt'):
    print("{}Reading the contents of config.txt file... \n{}".format(BOLD, ENDC))
    time.sleep(0.3)
    config = {}
    with open(config_file, 'r') as f:
        for line in f:
            line = line.strip()
            if '=' in line and not line.startswith('#'):
                key, value = line.split('=')
                key = key.strip()
                value = value.split('#')[0].strip()  # Remove comments and extra whitespace

                if key == "num_cores":
                    config[key] = int(value) if value else 4  # Default to 4 cores if empty
                elif key in ["ga_run", "ga_num_evals", "ga_pop_size", "ga_num_generations", "sw_max_its"]:
                    config[key] = int(value) if value else 10  # Default to 10 if empty
                elif key in ["grid_center", "npts"]:
                    config[key] = [int(float(x)) for x in value.split(",")] if value else [0, 0, 0]  # Default to zeros if empty
                elif key == "spacing":
                    config[key] = float(value) if value else 0.375  # Default to 0.375 if empty
                else:
                    config[key] = value  # Default to string
    
    # Print current settings at the start
    print("{}Current Docking Parameters:{}".format(FAIL, ENDC))
    print("{}---------------------------{}".format(FAIL, ENDC))
    sys.stdout.flush()
    for key, value in config.items():
        print("{0}: {1}".format(key, value))
        sys.stdout.flush()
    print("{}---------------------------{}\n".format(FAIL, ENDC))
    sys.stdout.flush()
    return config

def generate_dpf(receptor_file, ligand_file, config):
    dpf_filename = ligand_file.replace(".pdbqt", ".dpf")
    command = [
        sys.executable,
        path_dpf,
        "-l", ligand_file,
        "-r", receptor_file,
        "-o", dpf_filename,
        "-p", "ga_run=" + str(config["ga_run"]),
        "-p", "ga_num_evals=" + str(config["ga_num_evals"]),
        "-p", "ga_pop_size=" + str(config["ga_pop_size"]),
        "-p", "ga_num_generations=" + str(config["ga_num_generations"]),
        "-p", "sw_max_its=" + str(config["sw_max_its"])
    ]
    subprocess.call(command)
    return dpf_filename

def generate_gpf(receptor_file, config, ligand_atom_types):
    gpf_filename = receptor_file.replace(".pdbqt", ".gpf")
    command_params = [
        sys.executable,
        path_gpf,
        "-l", receptor_file,
        "-r", receptor_file,
        "-o", gpf_filename,
        "-p", "gridcenter=" + ",".join(map(str, config["grid_center"])),
        "-p", "spacing=" + str(config["spacing"]),
        "-p", "npts=" + ",".join(map(str, config["npts"])),
    ]

    # Run the command to generate the .gpf file
    subprocess.call(command_params)
    return gpf_filename


def print_autogrid_time():
    # Search for 'autogrid.log' in the current directory
    log_file_name = "autogrid.log"
    current_dir = os.getcwd()
    log_file_path = os.path.join(current_dir, log_file_name)

    # Check if the log file exists
    if not os.path.isfile(log_file_path):
        print("Error: autogrid.log file not found in the current directory.")
        return

    with open(log_file_path, 'r') as log_file:
        lines = log_file.readlines()
    
    # Initialize real_time as None
    real_time = None

    # Traverse the lines in reverse order to find the last occurrence of "Real="
    for line in reversed(lines):
        if "Real=" in line:
            # Extract the time value from the line
            real_time = line.split("Real=")[-1].split(',')[0].strip()
            break
    
    # Check if the time was found
    if real_time:
        # Output the message with the correct formatting
        print("{}Successfully generated map files with AutoGrid4 in {}{}".format(OKGREEN, real_time, ENDC))
    else:
        print("Error: Could not find 'Real=' in the log file.")

def scan_dpf_for_atom_types(dpf_filename, ligand_name):
    atom_types = set()
    try:
        with open(dpf_filename, 'r') as file:
            for line in file:
                line = line.strip()

                # Check if the line contains 'ligand_types'
                if 'ligand_types' in line:
                    parts = line.split('ligand_types', 1)
                    if len(parts) > 1:
                        value = parts[1].split('#', 1)[0].strip()  # Remove comments and extra spaces
                        types = value.split()
                        atom_types.update(types)
    except IOError:
        print("Error: File not found or cannot be opened: {0}".format(dpf_filename))

    # Print extracted atom types with the ligand name
    print("Extracted atom types: {0} from {1}".format(atom_types, ligand_name))
    sys.stdout.flush()
    return atom_types

def check_gpf_for_missing_maps(gpf_filename, ligand_atom_types):
    missing_maps = []
    try:
        with open(gpf_filename, 'r') as file:
            content = file.read()
            for atom_type in ligand_atom_types:
                if "map receptor.{0}.map".format(atom_type) not in content:
                    missing_maps.append(atom_type)
    except IOError:
        print("GPF file not found: {0}".format(gpf_filename))
    return missing_maps

def add_missing_maps_to_gpf(gpf_filename, missing_maps):
    lines = []
    try:
        with open(gpf_filename, 'r') as file:
            lines = file.readlines()
    except IOError:
        print("GPF file not found: {0}".format(gpf_filename))
        return
    with open(gpf_filename, 'w') as file:
        for line in lines:
            file.write(line)
            if line.startswith('smooth'):
                for atom_type in missing_maps:
                    file.write("map receptor.{0}.map\n".format(atom_type))

def add_missing_ligand_types_to_gpf(gpf_filename, ligand_atom_types):
    # Read existing GPF file content
    try:
        with open(gpf_filename, 'r') as file:
            lines = file.readlines()
    except IOError:
        print("GPF file not found: {0}".format(gpf_filename))
        return

    # Find the index of the 'ligand_types' line
    ligand_types_line_index = -1
    for i, line in enumerate(lines):
        if line.startswith("ligand_types"):
            ligand_types_line_index = i
            break
    if ligand_types_line_index == -1:
        print("ligand_types line not found in: {0}".format(gpf_filename))
        return

     # Extract the existing ligand types and identify missing ones
    existing_ligand_types = set()
    line_parts = lines[ligand_types_line_index].split()
    if len(line_parts) > 1:
        # Extract existing ligand types (excluding comment)
        comment_index = line_parts.index('#') if '#' in line_parts else len(line_parts)
        existing_ligand_types = set(line_parts[1:comment_index])

    # Determine missing atom types
    missing_atom_types = ligand_atom_types - existing_ligand_types
    if not missing_atom_types:
        print("No missing ligand types to add.")
        sys.stdout.flush()
        return

    # Update the 'ligand_types' line with missing atom types
    all_ligand_types = sorted(existing_ligand_types.union(ligand_atom_types))
    new_ligand_types_line = "ligand_types " + " ".join(all_ligand_types)

     # Preserve any existing comment on the line
    if '#' in lines[ligand_types_line_index]:
        comment = lines[ligand_types_line_index].split('#', 1)[1]
        new_ligand_types_line += " #" + comment.strip()
    new_ligand_types_line += "\n"

    # Write the updated content back to the file
    with open(gpf_filename, 'w') as file:
        for i, line in enumerate(lines):
            if i == ligand_types_line_index:
                file.write(new_ligand_types_line)
            else:
                file.write(line)
    print("{}Updated contents of ligand_types in {}{}".format(WARNING, gpf_filename, ENDC))
    sys.stdout.flush()

def run_autogrid(gpf_filename):
    print("{}Running AutoGrid4...{}".format(BOLD, ENDC))
    command = ["autogrid4", "-p", gpf_filename, "-l", "autogrid.log"]
    subprocess.call(command)

def run_autodock(dpf_filename, ligand_file):
    log_filename = ligand_file.replace(".pdbqt", ".dlg")
    command = ["autodock4", "-p", dpf_filename, "-l", log_filename]
    subprocess.call(command)

def process_ligand(args):
    ligand_file, receptor_file, config, completed_dockings, lock = args
    print("{}Initiating docking of ligand: {}{}".format(OKBLUE, ligand_file, ENDC))
    sys.stdout.flush()
    dpf_filename = generate_dpf(receptor_file, ligand_file, config)
    # Introduce a delay to reduce system strain
    time.sleep(5)
    run_autodock(dpf_filename, ligand_file)
    print("{}Completed the docking of ligand: {}{}".format(OKGREEN, ligand_file, ENDC))
    sys.stdout.flush()

    # Safely increment the shared completed_dockings counter
    with lock:  # Ensure that incrementing is thread-safe
        completed_dockings.value += 1



def main():
    try:
        # Prints the title
        title()

        time.sleep(2.3) # Shows title for 2.3 seconds

        userInterval()

        # Read the ligands, log and receptor files
        receptor_file = "receptor.pdbqt"
        ligand_files = [f for f in os.listdir(".") if f.endswith(".pdbqt") and f != receptor_file]
        # Read user configurations
        config = read_config('config.txt')

        # Use threading to run the keep_cmd_active function in the background
        wakecmd_thread = threading.Thread(target=keep_cmd_active, args=(interval_seconds,))
        wakecmd_thread.daemon = True  # Daemon thread will automatically exit when the main program exits
        wakecmd_thread.start()

        # Computes the number of ligand files (PDBQT)
        total_ligands = len(ligand_files)

        # Display number of ligands in workspace
        print("{}Total ligands to be docked: {}{}".format(WARNING, total_ligands, ENDC))
        sys.stdout.flush()

        # Use Manager to create shared objects
        manager = Manager()
        completed_dockings = manager.Value('i', 0)
        lock = manager.Lock()

        # Step 1: Generate DPF files, collect atom types and inform user
        print("{}Scanning for unique atom types in ligand PDBQT files to ensure AutoGrid4 correctly generates maps for each atom type{}".format(BOLD, ENDC))
        ligand_atom_types = set()
        for ligand_file in ligand_files:
            dpf_filename = generate_dpf(receptor_file, ligand_file, config)
            ligand_name = ligand_file.replace(".pdbqt", "")  # Extract ligand name without extension
            ligand_atom_types.update(scan_dpf_for_atom_types(dpf_filename, ligand_name))

        # Step 2: Generate GPF file
        gpf_filename = generate_gpf(receptor_file, config, ligand_atom_types)

        # Step 3: Check GPF file for missing maps and add them
        missing_maps = check_gpf_for_missing_maps(gpf_filename, ligand_atom_types)
        if missing_maps:
            add_missing_maps_to_gpf(gpf_filename, missing_maps)

        # Step 4: Add missing ligand types to GPF file
        add_missing_ligand_types_to_gpf(gpf_filename, ligand_atom_types)

        # Step 5: Run Autogrid
        run_autogrid(gpf_filename)

        # Step 6: Verify that all required maps were generated
        missing_maps = check_gpf_for_missing_maps(gpf_filename, ligand_atom_types)
        if missing_maps:
            # Re-run the map generation if necessary
            print("Missing maps detected, regenerating maps...")
            sys.stdout.flush()
            add_missing_maps_to_gpf(gpf_filename, missing_maps)
            run_autogrid(gpf_filename)
        
        # Display the autogrid runtime
        print_autogrid_time()

        # Display the status
        start_time = time.time()
        status_thread = threading.Thread(target=print_status, args=(start_time, total_ligands, completed_dockings))
        status_thread.daemon = True  # Daemonize thread to end it when the main program exits
        status_thread.start()

        # Prepare arguments for process_ligand
        args = [(ligand_file, receptor_file, config, completed_dockings, lock) for ligand_file in ligand_files]
        # Process ligands in parallel using apply_async
        num_cores = config.get("num_cores", 4) # Default is 4 cores, i.e. 4 docking at once
        pool = Pool(processes=num_cores)
        pool.map(process_ligand, args)
        pool.close()
        pool.join()

        print("{}Finished docking all ligands successfully!{}".format(OKGREEN, ENDC))
        sys.stdout.flush()
    
    except KeyboardInterrupt:
        print("Main script stopped.")
        exit(0)

if __name__ == '__main__':
    main()