import os
import glob
import subprocess
os.system('color')

# Define color codes
HEADER = '\033[95m'
OKBLUE = '\033[94m'
OKCYAN = '\033[96m'
OKGREEN = '\033[92m'
WARNING = '\033[93m'
FAIL = '\033[91m'
ENDC = '\033[0m'
BOLD = '\033[1m'
UNDERLINE = '\033[4m'

def convert_sdf_to_pdbqt():
    # Automatically detect the 'sdf_ligands' folder in the current directory
    current_dir = os.getcwd()
    sdf_folder = os.path.join(current_dir, "sdf_ligands")
    
    # Check if the 'sdf_ligands' folder exists
    if not os.path.exists(sdf_folder):
        print("Error: 'sdf_ligands' folder not found in the current directory.")
        return

    # Create the output folder in the current directory
    output_folder = os.path.join(current_dir, "ligand_pdbqt")
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    # Loop over all SDF files in the 'sdf_ligands' folder
    sdf_files = glob.glob(os.path.join(sdf_folder, "*.sdf"))
    
    if not sdf_files:
        print("No SDF files found in 'sdf_ligands' folder.")
        return
    
    successful_conversions = 0

    for sdf_file in sdf_files:
        base_name = os.path.basename(sdf_file)
        output_file = os.path.join(output_folder, base_name.replace(".sdf", ".pdbqt"))
        
        # Construct the OpenBabel command
        command = [
            'obabel',            # OpenBabel command-line tool
            sdf_file,            # Input SDF file
            '-O', output_file,  # Output PDBQT file
            '-h',                # Add hydrogens
            '--add-implicit-hydrogens',  # Add implicit hydrogens
            '-p',                # Compute partial charges
            'gasteiger'          # Use Gasteiger method for charges
        ]
        
        # Execute the command
        result = subprocess.call(command)
        
        # Extract file names for print statements
        input_file_name = os.path.basename(sdf_file)
        output_file_name = os.path.basename(output_file)
        
        if result == 0:
            print("Converted {0} to {1}".format(input_file_name, output_file_name))
            successful_conversions += 1
        else:
            print("Failed to convert {0}".format(input_file_name))
    
    # Print summary of successful conversions
    print("{}{} molecules converted successfully!{}".format(OKGREEN, successful_conversions, ENDC))

if __name__ == "__main__":
    convert_sdf_to_pdbqt()
